@@ -19,13 +19,12 @@ from .child import Child
 from .cli_stub import CLIOptions
 from .constants import appname, kitty_exe
 from .fast_data_types import (
-    add_tab, attach_window, detach_window, get_boss, mark_tab_bar_dirty,
-    next_window_id, remove_tab, remove_window, ring_bell, set_active_tab,
-    set_active_window, swap_tabs, sync_os_window_title
+    add_tab, attach_window, detach_window, get_boss, get_options,
+    mark_tab_bar_dirty, next_window_id, remove_tab, remove_window, ring_bell,
+    set_active_tab, set_active_window, swap_tabs, sync_os_window_title
 )
 from .layout.base import Layout, Rect
 from .layout.interface import create_layout_object_for, evict_cached_layouts
-from .options_stub import Options
 from .tab_bar import TabBar, TabBarData
 from .typing import EdgeLiteral, SessionTab, SessionType, TypedDict
 from .utils import log_error, platform_window_id, resolved_shell
@@ -90,10 +89,10 @@ class Tab:  # {{{
         self.id: int = add_tab(self.os_window_id)
         if not self.id:
             raise Exception('No OS window with id {} found, or tab counter has wrapped'.format(self.os_window_id))
-        self.opts, self.args = tab_manager.opts, tab_manager.args
+        self.args = tab_manager.args
         self.name = getattr(session_tab, 'name', '')
-        self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or self.opts.enabled_layouts]
-        self.borders = Borders(self.os_window_id, self.id, self.opts)
+        self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or get_options().enabled_layouts]
+        self.borders = Borders(self.os_window_id, self.id)
         self.windows = WindowList(self)
         for i, which in enumerate('first second third fourth fifth sixth seventh eighth ninth tenth'.split()):
             setattr(self, which + '_window', partial(self.nth_window, num=i))
@@ -116,6 +115,14 @@ class Tab:  # {{{
             self._set_current_layout(l0)
             self.startup(session_tab)
 
+    def apply_options(self) -> None:
+        for window in self:
+            window.apply_options()
+        self.enabled_layouts = [x.lower() for x in get_options().enabled_layouts] or ['tall']
+        if self.current_layout.name not in self.enabled_layouts:
+            self._set_current_layout(self.enabled_layouts[0])
+        self.relayout()
+
     def take_over_from(self, other_tab: 'Tab') -> None:
         self.name, self.cwd = other_tab.name, other_tab.cwd
         self.enabled_layouts = list(other_tab.enabled_layouts)
@@ -216,6 +223,7 @@ class Tab:  # {{{
         return create_layout_object_for(name, self.os_window_id, self.id)
 
     def next_layout(self) -> None:
+        '@ac:lay: Go to the next enabled layout'
         if len(self.enabled_layouts) > 1:
             for i, layout_name in enumerate(self.enabled_layouts):
                 if layout_name == self.current_layout.full_name:
@@ -228,11 +236,19 @@ class Tab:  # {{{
             self.relayout()
 
     def last_used_layout(self) -> None:
+        '@ac:lay: Go to the previously used layout'
         if len(self.enabled_layouts) > 1 and self._last_used_layout and self._last_used_layout != self._current_layout_name:
             self._set_current_layout(self._last_used_layout)
             self.relayout()
 
     def goto_layout(self, layout_name: str, raise_exception: bool = False) -> None:
+        '''
+        @ac:lay: Switch to the named layout
+
+        For example::
+
+            map f1 goto_layout tall
+        '''
         layout_name = layout_name.lower()
         if layout_name not in self.enabled_layouts:
             if raise_exception:
@@ -242,6 +258,21 @@ class Tab:  # {{{
         self._set_current_layout(layout_name)
         self.relayout()
 
+    def toggle_layout(self, layout_name: str) -> None:
+        '''
+        @ac:lay: Toggle the named layout
+
+        Switches to the named layout if another layout is current, otherwise
+        switches to the last used layout. Useful to "zoom" a window temporarily
+        by switching to the stack layout. For example::
+
+            map f1 toggle_layout stack
+        '''
+        if self._current_layout_name == layout_name:
+            self.last_used_layout()
+        else:
+            self.goto_layout(layout_name)
+
     def resize_window_by(self, window_id: int, increment: float, is_horizontal: bool) -> Optional[str]:
         increment_as_percent = self.current_layout.bias_increment_for_cell(is_horizontal) * increment
         if self.current_layout.modify_size_of_window(self.windows, window_id, increment_as_percent, is_horizontal):
@@ -250,6 +281,11 @@ class Tab:  # {{{
         return 'Could not resize'
 
     def resize_window(self, quality: str, increment: int) -> None:
+        '''
+        @ac:win: Resize the active window by the specified amount
+
+        See :ref:`window_resizing` for details.
+        '''
         if increment < 1:
             raise ValueError(increment)
         is_horizontal = quality in ('wider', 'narrower')
@@ -260,10 +296,12 @@ class Tab:  # {{{
             ring_bell()
 
     def reset_window_sizes(self) -> None:
+        '@ac:win:Reset window sizes undoing any dynamic resizing of windows'
         if self.current_layout.remove_all_biases():
             self.relayout()
 
     def layout_action(self, action_name: str, args: Sequence[str]) -> None:
+        '@ac:lay: Perform a layout specific action. See :doc:`layouts` for details'
         ret = self.current_layout.layout_action(action_name, args, self.windows)
         if ret is None:
             ring_bell()
@@ -283,15 +321,15 @@ class Tab:  # {{{
         check_for_suitability = True
         if cmd is None:
             if use_shell:
-                cmd = resolved_shell(self.opts)
+                cmd = resolved_shell(get_options())
                 check_for_suitability = False
             else:
                 if self.args.args:
                     cmd = list(self.args.args)
                 else:
-                    cmd = resolved_shell(self.opts)
+                    cmd = resolved_shell(get_options())
                     check_for_suitability = False
-                cmd = self.args.args or resolved_shell(self.opts)
+                cmd = self.args.args or resolved_shell(get_options())
         if check_for_suitability:
             old_exe = cmd[0]
             if not os.path.isabs(old_exe):
@@ -310,7 +348,7 @@ class Tab:  # {{{
                 else:
                     if stat.S_ISDIR(st.st_mode):
                         cwd = old_exe
-                        cmd = resolved_shell(self.opts)
+                        cmd = resolved_shell(get_options())
                     elif not is_executable:
                         import shlex
                         with suppress(OSError):
@@ -320,7 +358,7 @@ class Tab:  # {{{
                                     line = f.read(4096).splitlines()[0]
                                     cmd += shlex.split(line) + [old_exe]
                                 else:
-                                    cmd += [resolved_shell(self.opts)[0], cmd[0]]
+                                    cmd += [resolved_shell(get_options())[0], cmd[0]]
         fenv: Dict[str, str] = {}
         if env:
             fenv.update(env)
@@ -328,7 +366,7 @@ class Tab:  # {{{
         pwid = platform_window_id(self.os_window_id)
         if pwid is not None:
             fenv['WINDOWID'] = str(pwid)
-        ans = Child(cmd, cwd or self.cwd, self.opts, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
+        ans = Child(cmd, cwd or self.cwd, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
         ans.fork()
         return ans
 
@@ -356,7 +394,7 @@ class Tab:  # {{{
         child = self.launch_child(
             use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env, allow_remote_control=allow_remote_control)
         window = Window(
-            self, child, self.opts, self.args, override_title=override_title,
+            self, child, self.args, override_title=override_title,
             copy_colors_from=copy_colors_from, watchers=watchers
         )
         # Must add child before laying out so that resize_pty succeeds
@@ -386,11 +424,13 @@ class Tab:  # {{{
         )
 
     def close_window(self) -> None:
+        '@ac:win: Close the currently active window'
         w = self.active_window
         if w is not None:
             self.remove_window(w)
 
     def close_other_windows_in_tab(self) -> None:
+        '@ac:win: Close all windows in the tab other than the currently active window'
         if len(self.windows) > 1:
             active_window = self.active_window
             for window in tuple(self.windows):
@@ -429,6 +469,13 @@ class Tab:  # {{{
             return self.current_layout.nth_window(self.windows, n)
 
     def nth_window(self, num: int = 0) -> None:
+        '''
+        @ac:win: Focus the nth window if positive or the previously active windows if negative
+
+        For example, to ficus the previously active window::
+
+            map ctrl+p nth_window -1
+        '''
         if self.windows:
             if num < 0:
                 self.windows.make_previous_group_active(-num)
@@ -442,9 +489,11 @@ class Tab:  # {{{
             self.relayout_borders()
 
     def next_window(self) -> None:
+        '@ac:win: Focus the next window in the current tab'
         self._next_window()
 
     def previous_window(self) -> None:
+        '@ac:win: Focus the previous window in the current tab'
         self._next_window(-1)
 
     prev_window = previous_window
@@ -467,11 +516,27 @@ class Tab:  # {{{
             return self.most_recent_group(candidates)
 
     def neighboring_window(self, which: EdgeLiteral) -> None:
+        '''
+        @ac:win: Focus the neighboring window in the current tab
+
+        For example::
+
+            map ctrl+left neighboring_window left
+            map ctrl+down neighboring_window bottom
+        '''
         neighbor = self.neighboring_group_id(which)
         if neighbor:
             self.windows.set_active_group(neighbor)
 
     def move_window(self, delta: Union[EdgeLiteral, int] = 1) -> None:
+        '''
+        @ac:win: Move the window in the specified direction
+
+        For example::
+
+            map ctrl+left move_window left
+            map ctrl+down move_window bottom
+        '''
         if isinstance(delta, int):
             if self.current_layout.move_window(self.windows, delta):
                 self.relayout()
@@ -482,14 +547,17 @@ class Tab:  # {{{
                     self.relayout()
 
     def move_window_to_top(self) -> None:
+        '@ac:win: Move active window to the top (make it the first window)'
         n = self.windows.active_group_idx
         if n > 0:
             self.move_window(-n)
 
     def move_window_forward(self) -> None:
+        '@ac:win: Move active window forward (swap it with the next window)'
         self.move_window()
 
     def move_window_backward(self) -> None:
+        '@ac:win: Move active window backward (swap it with the previous window)'
         self.move_window(-1)
 
     def list_windows(self, active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[WindowDict, None, None]:
@@ -509,6 +577,10 @@ class Tab:  # {{{
     def __len__(self) -> int:
         return len(self.windows)
 
+    @property
+    def num_window_groups(self) -> int:
+        return self.windows.num_groups
+
     def __contains__(self, window: Window) -> bool:
         return window in self.windows
 
@@ -530,16 +602,16 @@ class Tab:  # {{{
 
 class TabManager:  # {{{
 
-    def __init__(self, os_window_id: int, opts: Options, args: CLIOptions, wm_class: str, wm_name: str, startup_session: Optional[SessionType] = None):
+    def __init__(self, os_window_id: int, args: CLIOptions, wm_class: str, wm_name: str, startup_session: Optional[SessionType] = None):
         self.os_window_id = os_window_id
         self.wm_class = wm_class
         self.wm_name = wm_name
         self.last_active_tab_id = None
-        self.opts, self.args = opts, args
-        self.tab_bar_hidden = self.opts.tab_bar_style == 'hidden'
+        self.args = args
+        self.tab_bar_hidden = get_options().tab_bar_style == 'hidden'
         self.tabs: List[Tab] = []
         self.active_tab_history: Deque[int] = deque()
-        self.tab_bar = TabBar(self.os_window_id, opts)
+        self.tab_bar = TabBar(self.os_window_id)
         self._active_tab_idx = 0
 
         if startup_session is not None:
@@ -581,7 +653,7 @@ class TabManager:  # {{{
 
     @property
     def tab_bar_should_be_visible(self) -> bool:
-        return len(self.tabs) >= self.opts.tab_bar_min_tabs
+        return len(self.tabs) >= get_options().tab_bar_min_tabs
 
     def _add_tab(self, tab: Tab) -> None:
         visible_before = self.tab_bar_should_be_visible
@@ -765,16 +837,16 @@ class TabManager:  # {{{
 
         if active_tab_needs_to_change:
             next_active_tab = -1
-            if self.opts.tab_switch_strategy == 'previous':
+            if get_options().tab_switch_strategy == 'previous':
                 while self.active_tab_history and next_active_tab < 0:
                     tab_id = self.active_tab_history.pop()
                     for idx, qtab in enumerate(self.tabs):
                         if qtab.id == tab_id:
                             next_active_tab = idx
                             break
-            elif self.opts.tab_switch_strategy == 'left':
+            elif get_options().tab_switch_strategy == 'left':
                 next_active_tab = max(0, self.active_tab_idx - 1)
-            elif self.opts.tab_switch_strategy == 'right':
+            elif get_options().tab_switch_strategy == 'right':
                 next_active_tab = min(self.active_tab_idx, len(self.tabs) - 1)
 
             if next_active_tab < 0:
@@ -806,7 +878,7 @@ class TabManager:  # {{{
                     has_activity_since_last_focus = True
             ans.append(TabBarData(
                 title, t is at, needs_attention,
-                len(t), t.current_layout.name or '',
+                len(t), t.num_window_groups, t.current_layout.name or '',
                 has_activity_since_last_focus
             ))
         return ans
@@ -829,4 +901,13 @@ class TabManager:  # {{{
         self.tab_bar.destroy()
         del self.tab_bar
         del self.tabs
+
+    def apply_options(self) -> None:
+        for tab in self:
+            tab.apply_options()
+        self.tab_bar_hidden = get_options().tab_bar_style == 'hidden'
+        self.tab_bar.apply_options()
+        self.update_tab_bar_data()
+        self.mark_tab_bar_dirty()
+        self.tab_bar.layout()
 # }}}
